#!/usr/bin/env python3

# 递归搜索目录下的所有 .cbp 文件，并生成对应的 Makefile 文件
# 如果有多个 .cbp 文件，则在根目录放一个

# 同时生成一个  .vscode/task.json 的配置，以及 winmk.bat
# winmk.bat:
# SET SCRIPT_PATH=%~dp0%
# SET PATH=%SCRIPT_PATH%\..\tools\utils;%PATH%
# make "%1" -j %NUMBER_OF_PROCESSOR%

# task.json
# {
#     "version": "2.0.0",
#     "tasks": [
#         {
#             "label": "all",
#             "type": "shell",
#             "command": ".vscode/winmk.bat all",
#             "problemMatcher": [],
#             "group": {
#                 "kind": "build",
#                 "isDefault": true
#             }
#         },
#         {
#             "label": "clean",
#             "type": "shell",
#             "command": ".vscode/winmk.bat clean",
#             "problemMatcher": [],
#             "group": "build"
#         }
#     ]
# }

import os
import sys
import argparse
import subprocess
import pathlib

def find_all_cbp_files(_dir):
    cbp_files = []
    for root, dirs, files in os.walk(_dir):
        for f in files:
            if f.endswith('.cbp'):
                cbp_files.append(os.path.join(root, f))

    return cbp_files

def write_task_body(fout, target):
    fout.write('            "label": "{target}",\n'.format(target=target))
    fout.write('            "type": "shell",\n')
    fout.write('            "windows": {\n')
    fout.write('                "command": ".vscode/winmk.bat {target}"\n'.format(target=target))
    fout.write('            },\n')
    fout.write('            "command": "make {target} -j`nproc`",\n'.format(target=target))
    fout.write('            "problemMatcher": [],\n')

def write_task(fout, target):
    fout.write('        ,{\n')
    write_task_body(fout, target)
    fout.write('            "group": "build"\n')
    fout.write('        }\n')

def create_vscode_conf(projdir, targets, clean_targets):
    # and we should create .vscode configs
    vsc = os.path.join(projdir, '.vscode')
    os.makedirs(vsc, exist_ok=True)

    # 1. create winmk.bat
    with open(os.path.join(vsc, 'winmk.bat'), 'w') as fout:
        fout.write('SET SCRIPT_PATH=%~dp0%\r\n')
        fout.write('SET PATH=%SCRIPT_PATH%\\..\\tools\\utils;%PATH%\r\n')
        fout.write('make "%1" -j %NUMBER_OF_PROCESSORS%\r\n')

    # 2. create tasks.jon
    with open(os.path.join(vsc, 'tasks.json'), 'w') as fout:
        fout.write('{\n')
        fout.write('    "version": "2.0.0",\n')
        fout.write('    "tasks": [\n')
        fout.write('        {\n')
        write_task_body(fout, "all")
        fout.write('            "group": {\n')
        fout.write('                "kind": "build",\n')
        fout.write('                "isDefault": true\n')
        fout.write('            }\n')
        fout.write('        }\n')

        write_task(fout, 'clean')
        
        for t in targets:
            write_task(fout, t)
        for t in clean_targets:
            write_task(fout, t)

        fout.write('    ]\n')
        fout.write('}\n')

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Generate Makefile for CBP')
    parser.add_argument('--dir', type=str, default='', help='directory')
    parser.add_argument('--cbp2make', type=str, default='', help='path to cbp2make')
    parser.add_argument('--cbp2vscode', type=str, default='/opt/utils/cbp2vscode', help='path to cbp2vscode')

    args = parser.parse_args(sys.argv[1:])

    # 1. find all cbp files
    projdir = os.path.abspath(args.dir)

    cbp_files = find_all_cbp_files(projdir)

    if len(cbp_files) == 0:
        sys.exit(0)

    subtargets = []

    cbp_dirs = {}
    for c in cbp_files:
        cbp_dir = os.path.dirname(c)
        if cbp_dir not in cbp_dirs:
            cbp_dirs[cbp_dir] = 1
        else:
            cbp_dirs[cbp_dir] += 1

    if len(cbp_files) == 1:
        # only one .cbp, check if in the root dir
        cbp_file = cbp_files[0]
        cbp_dir = os.path.dirname(cbp_file)

        makefile = os.path.join(cbp_dir, 'Makefile')

        if cbp_dir == projdir:
            # only one .cbp in projdir, we just create an Makefile
            subprocess.call([args.cbp2make, '--cbp', cbp_file, '--output', makefile])
        else:
            # not in projdir, we should create an Makefile in the same dir
            subprocess.call([args.cbp2make, '--cbp', cbp_file, '--output', makefile])
            # and create an Makefile in root, 
            subtargets.append([cbp_file, makefile])
    else:
        for cbp_file in cbp_files:
            cbp_dir = os.path.dirname(cbp_file)
            cbp_title = os.path.basename(cbp_file)[:-4].lower()
            build_dir = 'objs'
            if cbp_dirs[cbp_dir] > 1:
                makefile = os.path.join(cbp_dir, 'Makefile.' + cbp_title)
                build_dir = cbp_title + '_objs'
            else:
                if cbp_dir == projdir:
                    makefile = os.path.join(cbp_dir, 'Makefile.' + cbp_title)
                    build_dir = cbp_title + '_objs'
                else:
                    makefile = os.path.join(cbp_dir, 'Makefile')
            
            subprocess.call([args.cbp2make, '--cbp', cbp_file, '--output', makefile, '--builddir', build_dir])
            subtargets.append([cbp_file, makefile])

    if len(subtargets) != 0:
        # we should create Makefile in projdir
        rootmakefile = os.path.join(projdir, 'Makefile')

        # we should genrate default.workspace in root
        with open(os.path.join(projdir, 'default.workspace'), 'w') as fout:
            fout.write('<?xml version="1.0" encoding="UTF-8" standalone="yes" ?>\n')
            fout.write('<CodeBlocks_workspace_file>\n')
            fout.write('\t<Workspace title="Workspace">\n')
            for st in subtargets:
                cbp_file, _ = st
                rela_path = str(pathlib.Path(cbp_file).relative_to(projdir))
                fout.write('\t\t<Project filename="{}" />\n'.format(rela_path))
            fout.write('\t</Workspace>\n')
            fout.write('</CodeBlocks_workspace_file>\n')

        lines = []

        target_lines = []
        targets = []
        clean_targets = []

        lines.append('# 总的 Makefile，用于调用目录下各个子工程对应的 Makefile')
        lines.append("""# 注意： Linux 下编译方式：
# 1. 从 http://pkgman.jieliapp.com/doc/all 处找到下载链接
# 2. 下载后，解压到 /opt/jieli 目录下，保证
#   /opt/jieli/common/bin/clang 存在（注意目录层次）
# 3. 确认 ulimit -n 的结果足够大（建议大于8096），否则链接可能会因为打开文件太多而失败
#   可以通过 ulimit -n 8096 来设置一个较大的值""")

        for st in subtargets:
            cbp_file, makefile = st
            cbp_dir = os.path.dirname(cbp_file)
            if cbp_dir == projdir:
                cbp_dir = '.'
            else:
                cbp_dir = cbp_dir[len(projdir)+1:]
            makefile = os.path.basename(makefile)
            # make -C {cbp_dir} -f {makefile}
            cbp_title = os.path.basename(cbp_file)[:-4].lower() # remove .cbp
            cbp_title = cbp_title.replace(' ', '_')

            targets.append(cbp_title)
            clean_targets.append('clean_' + cbp_title)

            target_lines.append(cbp_title + ':')
            target_lines.append('\t$(MAKE) -C ' + cbp_dir + ' -f ' + makefile)
            target_lines.append('')

            target_lines.append('clean_' + cbp_title + ':')
            target_lines.append('\t$(MAKE) -C ' + cbp_dir + ' -f ' + makefile + ' clean')
            target_lines.append('')

        lines.append('# 支持的目标')
        for t in targets:
            lines.append('# make {}'.format(t))
        lines.append('')

        lines.append('.PHONY: all clean {}'.format(' '.join(targets + clean_targets)))
        lines.append('')
        lines.append('all: {}'.format(' '.join(targets)))
        lines.append('\t@echo +ALL DONE')
        lines.append('')

        lines.append('clean: {}'.format(' '.join(clean_targets)))
        lines.append('\t@echo +CLEAN DONE\n')
        lines.append('')

        with open(rootmakefile, 'w') as fout:
            fout.write('\n'.join(lines))
            fout.write('\n'.join(target_lines))

        create_vscode_conf(projdir, targets, clean_targets)
    else:
        create_vscode_conf(projdir, [], [])

    if args.cbp2vscode != '':
        # create .vscode/c_cpp_properites.json
        for cbp in cbp_files:
            subprocess.call([args.cbp2vscode, '--cbp', cbp, '--root', projdir])
